package com.sillycycle.bagleyd.rubik;

/*
 * @(#)Rubik2D.java
 *
 * Copyright 1994 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;

/**
 * The <code>Rubik2D</code> class does all the 2D graphics for
 * <code>RubikFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class Rubik2D extends Canvas {
	private static final long serialVersionUID = 42L;
	static final int FRONT_LOC = 0;
	static final int BACK_LOC = 11;
	static final int COORD2D = 8;
	static final int MAXX = 3;
	static final int MAXY = 4;
	static final int MAXXY = 4; /* (MAX(MAXX,MAXY)) */
	static final int MAX_RECT = 12; /* (MAXX*MAXY) */

	transient RubikNextRow[] rotateToRow = /*CW to min face */
		new RubikNextRow[Rubik.MAX_FACES];
	static final int[/*MAX_RECT*/] planeToCube =
		{6, 0, 6, 1, 2, 3, 6, 4, 6, 6, 5, 6};
	static int[/*Rubik.MAX_FACES*/] cubeToPlane =
		{1, 3, 4, 5, 7, 10};

	Point pos = new Point(0, 0);
	Point puzzleSize, coreSize, puzzleOffset;
	int cubeLength, delta, orientLineLength;
	int viewLengthX, viewLengthY, viewLengthZ;
	int faceLengthX, faceLengthY, faceLengthZ;
	int lastX, lastY;

	JFrame      frame = null;
	Rubik       rubik = null;
	int         randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false, vertical = false;

	public Rubik2D(JFrame frame, Rubik rubik) {
		this.frame = frame;
		this.rubik = rubik;
		firstPaint = true;
	}

	public boolean checkStarted() {
		return rubik.checkStarted();
	}

	void facePosition(int faceX, int faceY) {
		int i;

		pos.x = puzzleOffset.x;
		pos.y = puzzleOffset.y;
		for (i = 0; i < faceX; i++) {
			if ((i & 1) == 1) {
				pos.x += viewLengthX;
			} else {
				pos.x += viewLengthZ;
			}
		}
		for (i = 0; i < faceY; i++) {
			if ((i & 1) == 1) {
				pos.y += viewLengthY;
			} else {
				pos.y += viewLengthZ;
			}
		}
	}

	void drawOrientLine(Graphics g, int rotation, int dx, int dy) {
		int x1 = dx, x2 = dx, y1 = dy, y2 = dy;

		switch (rotation) {
		case Rubik.TOP:
			x1 = x2 = dx + cubeLength / 2;
			y2 += orientLineLength;
			break;
		case Rubik.RIGHT:
			y1 = y2 = dy + cubeLength / 2;
			x1 += cubeLength;
			x2 += cubeLength - orientLineLength - 1;
			break;
		case Rubik.BOTTOM:
			x1 = x2 = dx + cubeLength / 2;
			y1 += cubeLength;
			y2 += cubeLength - orientLineLength - 1;
			break;
		case Rubik.LEFT:
			y1 = y2 = dy + cubeLength / 2;
			x2 += orientLineLength;
			break;
		default:
			System.out.println("drawOrientLine: rotation " + rotation);
		}
		g.drawLine(x1, y1, x2, y2);
	}

	void fillRect(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY) {

		g.setColor(fillColor);
		g.fillRect(startX, startY, cubeLength, cubeLength);
		g.setColor(lineColor);
		g.drawRect(startX, startY, cubeLength, cubeLength);
	}

	void drawSquare(int face, int position, int offsetPosition) {
		Graphics g = rubik.getGraphics();

		if (g == null)
			return;
		try {
			int rotation, i, j, l;
			int sizeOfRow = rubik.sizeRow(face);
			int sizeOfColumn = rubik.sizeColumn(face);
			Point start = new Point(0, 0);
			Color fillColor, lineColor;

			i = position % sizeOfRow;
			j = position / sizeOfRow;
			rotation = rubik.cubeLoc[face][position].rotation;
			if (vertical || face != Rubik.MAX_FACES - 1) {
				facePosition(cubeToPlane[face] % MAXX,
					cubeToPlane[face] / MAXX);
					// returns pos
				start.x = pos.x + i * (cubeLength + delta - 1);
				start.y = pos.y + j * (cubeLength + delta - 1);
			} else {
				facePosition(cubeToPlane[face] / MAXX,
					cubeToPlane[face] % MAXX);
					// returns pos
				start.x = pos.x + (sizeOfRow - 1 - i) *
					(cubeLength + delta - 1);
				start.y = pos.y + (sizeOfColumn - 1 - j) *
					(cubeLength + delta - 1);
				rotation = (rotation + Rubik.HALF) % Rubik.STRT;
			}
			start.x += delta - 1;
			start.y += delta - 1;
			l = rubik.cubeLoc[face][position].face;
			if (offsetPosition == 0) {
				if (rubik.mono) {
					if (rubik.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = rubik.faceColor[l];
					lineColor = rubik.borderColor;
				}
			} else {
				if (rubik.mono) {
					if (rubik.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = rubik.faceColor[l];
					fillColor = rubik.borderColor;
				}
			}
			fillRect(g, fillColor, lineColor, start.x, start.y);
			if (rubik.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = cubeLength / 2;
				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(rubik.faceNames.charAt(l));
				g.drawString(buf,
					start.x + cubeLength / 3,
					start.y + 2 * cubeLength / 3);
			}
			if (rubik.orientChoice) {
				g.setColor(lineColor);
				drawOrientLine(g, rotation, start.x, start.y);
			}
		} finally {
			g.dispose();
		}
	}

	void drawFace(int face) {
		for (int position = 0; position < rubik.sizeFace(face); position++)
			drawSquare(face, position, 0);
	}

	void drawAllPieces() {
		for (int face = 0; face < Rubik.MAX_FACES; face++)
			drawFace(face);
	}

	void drawFrame(boolean show) {
		Graphics g = rubik.getGraphics();

		if (g == null)
			return;
		try {
			if (show) {
				Point tempLength;
				Point[] coord = new Point[MAXXY + 1];
				Point letters;
				int fontSize = (cubeLength + faceLengthZ) / 5;

				tempLength = new Point(puzzleOffset.x, puzzleOffset.y);
				for (int i = 0; i <= MAXXY; i++) {
					coord[i] = new Point(tempLength.x, tempLength.y);
					if ((i & 1) == 1) {
						tempLength.x += viewLengthX;
						tempLength.y += viewLengthY;
					} else {
						tempLength.x += viewLengthZ;
						tempLength.y += viewLengthZ;
					}
				}

				g.setFont(new Font("arial", 0, fontSize));
				g.setColor(rubik.currentForeground);
				g.drawLine(coord[1].x, coord[0].y, coord[2].x, coord[0].y);
				g.drawLine(coord[3].x, coord[1].y, coord[3].x, coord[2].y);
				g.drawLine(coord[1].x, coord[3].y, coord[2].x, coord[3].y);
				g.drawLine(coord[0].x, coord[1].y, coord[0].x, coord[2].y);
				letters = new Point(
					(coord[0].x + coord[1].x) / 2 - delta,
					(coord[0].y + coord[1].y) / 2);
				g.setColor(rubik.foreground);
				g.drawString("Front",
					letters.x - 5 * fontSize / 4,
					letters.y + 2 * fontSize / 5);
				letters.x = (coord[2].x + coord[3].x) / 2 - delta;
				letters.y = (coord[2].y + coord[3].y) / 2;
				g.drawString("Back",
					letters.x - fontSize,
					letters.y + 2 * fontSize / 5);
				g.setColor(rubik.currentForeground);
				if (vertical) {
					g.drawLine(coord[1].x, coord[0].y, coord[1].x, coord[4].y);
					g.drawLine(coord[2].x, coord[0].y, coord[2].x, coord[4].y);
					g.drawLine(coord[0].x, coord[1].y, coord[3].x, coord[1].y);
					g.drawLine(coord[0].x, coord[2].y, coord[3].x, coord[2].y);
					g.drawLine(coord[1].x, coord[4].y, coord[2].x, coord[4].y);
				} else {
					g.drawLine(coord[0].x, coord[1].y, coord[4].x, coord[1].y);
					g.drawLine(coord[0].x, coord[2].y, coord[4].x, coord[2].y);
					g.drawLine(coord[1].x, coord[0].y, coord[1].x, coord[3].y);
					g.drawLine(coord[2].x, coord[0].y, coord[2].x, coord[3].y);
					g.drawLine(coord[4].x, coord[1].y, coord[4].x, coord[2].y);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(rubik.background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void resizePieces() {
		/* offsetList[DOWN] = new Point(0, 0);
		offsetList[UP] = new Point(cubeLength + 2, cubeLength + 2);
		letterList[DOWN] = new Point(cubeLength / 4 - 3,
			cubeLength / 4 + 5);
		letterList[UP] = new Point(3 * cubeLength / 4,
			3 * cubeLength / 4 + 5); */
		orientLineLength = cubeLength / 4;
	}

	void narrowSelection() {
		// sets rubik.currentFace, rubik.currentPosition, rubik.currentDirection
		if (rubik.currentFace == Rubik.MAX_FACES - 1 &&
				rubik.currentDirection < RubikInterface.MAX_ORIENT && !vertical)
			rubik.currentDirection = (rubik.currentDirection + Rubik.HALF) %
				RubikInterface.MAX_ORIENT;
		/* Remap to row movement */
		if (rubik.currentDirection == Rubik.CW || rubik.currentDirection == Rubik.CCW) {
			int i, j;
			int newFace = rotateToRow[rubik.currentFace].face;
			int sizeOfRow = rubik.sizeRow(newFace);
			int sizeOfColumn = rubik.sizeColumn(newFace);

			rubik.currentDirection = (rubik.currentDirection == Rubik.CCW) ?
				(rotateToRow[rubik.currentFace].direction + 2) % RubikInterface.MAX_ORIENT :
				rotateToRow[rubik.currentFace].direction;
			if (rotateToRow[rubik.currentFace].sideFace == Rubik.LEFT ||
				rotateToRow[rubik.currentFace].sideFace == Rubik.BOTTOM) {
				i = sizeOfRow - 1;
				j = sizeOfColumn - 1;
			} else {
				i = j = 0;
			}
			rubik.currentFace = newFace;
			rubik.currentPosition = j * sizeOfRow + i;
		}
	}

	int positionToPiece(int positionX, int positionY) {
		// sets rubik.currentPosition
		int faceX, faceY, i, j, face, tempLength;
		int x = positionX - puzzleOffset.x;
		int y = positionY - puzzleOffset.y;

		/* faceX = x / viewLength; */
		for (faceX = 0; faceX <= MAXY; faceX++) {
			if ((faceX & 1) == 1)
				tempLength = viewLengthX;
			else
				tempLength = viewLengthZ;
			if (x <= tempLength)
				break;
			x -= tempLength;
		}
		/* faceY = y / viewLength; */
		for (faceY = 0; faceY <= MAXY; faceY++) {
			if ((faceY & 1) == 1)
				tempLength = viewLengthY;
			else
				tempLength = viewLengthZ;
			if (y <= tempLength)
				break;
			y -= tempLength;
		}
		i = java.lang.Math.max((x - delta) /
			(cubeLength + delta - 1), 0);
		j = java.lang.Math.max((y - delta) /
			(cubeLength + delta - 1), 0);
		if ((faceX != 1 && faceY != 1) ||
				(faceX >= 3 && vertical) ||
				(faceY >= 3 && !vertical))
			return -1;
		if (faceX == 3)
			face = Rubik.MAX_FACES - 1;
		else
			face = planeToCube[faceX + faceY * MAXX];
		int sizeOfRow = rubik.sizeRow(face);
		int sizeOfColumn = rubik.sizeColumn(face);
		if (i >= sizeOfRow)
			i = sizeOfRow - 1;
		if (j >= sizeOfColumn)
			j = sizeOfColumn - 1;
		if (faceX == 3) {
			i = sizeOfRow - 1 - i;
			j = sizeOfColumn - 1 - j;
		}
		rubik.currentPosition = j * sizeOfRow + i;
		return face;
	}

	void resizePuzzle() {
		Point flatlength = new Point(rubik.sizeX + 2 * rubik.sizeZ,
			2 * rubik.sizeY + 2 * rubik.sizeZ);
		Point flatlengthh = new Point(2 * rubik.sizeX + 2 * rubik.sizeZ,
			rubik.sizeY + 2 * rubik.sizeZ);
		int tempLength;

		coreSize = new Point(rubik.getSize().width,
			rubik.getSize().height);
		delta = 4;
		vertical = (coreSize.y / flatlength.y +
			coreSize.x / flatlength.x <=
			coreSize.y / flatlengthh.y +
			coreSize.x / flatlengthh.x);
		if (!vertical)
			flatlength = flatlengthh;
		tempLength = java.lang.Math.min((coreSize.y - 2 * delta - 2) / flatlength.y,
			(coreSize.x - 2 * delta - 2) / flatlength.x);
		cubeLength = java.lang.Math.max(tempLength - delta + 1, 0);
		faceLengthX = rubik.sizeX * (cubeLength + delta - 1);
		faceLengthY = rubik.sizeY * (cubeLength + delta - 1);
		faceLengthZ = rubik.sizeZ * (cubeLength + delta - 1);
		viewLengthX = faceLengthX + delta - 1;
		viewLengthY = faceLengthY + delta - 1;
		viewLengthZ = faceLengthZ + delta - 1;
		if (vertical) {
			puzzleSize = new Point(viewLengthX - 1 +
				2 * (viewLengthZ - 1) + delta,
				2 * (viewLengthY - 1) +
				2 * (viewLengthZ - 1) + delta);
		} else {
			puzzleSize = new Point(2 * (viewLengthX - 1) +
				2 * (viewLengthZ - 1) + delta,
				viewLengthY - 1 +
				2 * (viewLengthZ - 1) + delta);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		rotateToRow[0] = new RubikNextRow(1, Rubik.LEFT, Rubik.TOP);
		rotateToRow[1] = new RubikNextRow(0, Rubik.BOTTOM, Rubik.RIGHT);
		rotateToRow[2] = new RubikNextRow(0, Rubik.RIGHT, Rubik.BOTTOM);
		rotateToRow[3] = new RubikNextRow(0, Rubik.TOP, Rubik.LEFT);
		rotateToRow[4] = new RubikNextRow(1, Rubik.RIGHT, Rubik.BOTTOM);
		rotateToRow[5] = new RubikNextRow(0, Rubik.LEFT, Rubik.TOP);
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		rubik.drawAllPieces();
	}

	void selectPuzzle(boolean control, int x, int y) {
		if ((rubik.currentFace = positionToPiece(x, y)) != -1) { //sets currentPosition
			if (control || rubik.practiceChoice || !rubik.checkSolved(rubik.orientChoice)) {
				selectPaint = true;
				rubik.repaint();
				return;
			}
			rubik.currentFace = -1;
		}
		rubik.currentFace = -1;
	}

	void releasePuzzle(boolean control, int x, int y) {
		if (rubik.currentFace == -1)
			return;
		rubik.nextControl = control;
		int tempPosition = rubik.currentPosition;
		rubik.nextFace = positionToPiece(x, y); // sets currentPosition
		rubik.nextPosition = rubik.currentPosition;
		rubik.currentPosition = tempPosition;
		releasePaint = true;
		rubik.repaint();
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			rubik.initializePuzzle();
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (rubik.currentFace != -1) {
				drawSquare(rubik.currentFace,
					rubik.currentPosition, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (rubik.currentFace != -1) {
				drawSquare(rubik.currentFace,
					rubik.currentPosition, 0);
				rubik.selectPieces();
				rubik.currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			if (rubik.currentControl)
				rubik.moveControlCb(rubik.currentFace,
					rubik.currentDirection);
			else {
				rubik.movePieces(rubik.currentFace,
					rubik.currentPosition,
					rubik.currentDirection);
				if (stackPaint && rubik.checkSolved(rubik.orientChoice)) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_SOLVED,
						2);
				}

			}
			rubik.currentDirection = rubik.currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(rubik.delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			rubik.randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public void keyPressed(KeyEvent event) {
		int modifiers = event.getModifiersEx();

		switch (event.getKeyCode()) {
		case KeyEvent.VK_DIVIDE:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.CCW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
		case KeyEvent.VK_NUMPAD8:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.TOP,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
		case KeyEvent.VK_NUMPAD4:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.LEFT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_CLEAR:
		case KeyEvent.VK_NUMPAD5:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.CW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
		case KeyEvent.VK_NUMPAD6:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.RIGHT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
		case KeyEvent.VK_NUMPAD2:
			rubik.movePuzzleInput(rubik.lastX, rubik.lastY,
				Rubik.BOTTOM,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
			break;
		default:
		}
	}
}
